use clap::{Args, ValueEnum};
use image::{DynamicImage, io::Reader as ImageReader};
use color_full16::{RgbaImage, RgbImage};
use std::io::Cursor;
use tool_helper::{Error, Input, Output};
use types::{Header, PSXImageConverter};

mod types;
mod color_full16;

#[derive(Copy, Clone, PartialEq, Eq, PartialOrd, Ord, ValueEnum)]
pub enum ColorType{
    Clut4,
    Clut8,
    Full16,
}

#[derive(Args)]
pub struct Arguments {
    #[clap(arg_enum, value_parser)]
    color_depth: ColorType
}


fn encode<T: PSXImageConverter>(image: T, mut output: Output) -> Result<(), Error> {
    let width  = image.width();
    let height = image.height();
    let header = Header::encode(width, height, 0, 0).ok_or(Error::from_callback(|| {format!("Image size {}, {} needs to be even", width, height)}))?;

    tool_helper::raw::write_generic(&mut output, header)?;
    for color in image {
        tool_helper::raw::write_generic(&mut output, color)?;
    }

    Ok(())
}

fn convert_full16(input: Input, output: Output) -> Result<(), Error> {
    match ImageReader::new(Cursor::new(tool_helper::input_to_vec(input)?)).with_guessed_format()?.decode() {
        Ok(image) => {
            match image {
                DynamicImage::ImageRgb8(image)  => encode(RgbImage::new(image), output),
                DynamicImage::ImageRgba8(image) => encode(RgbaImage::new(image), output),

                _ => Err(Error::from_str("Only RGB and RGBA images are supported for 16bit encoding"))
            }
        },
        Err(error) => Err(Error::from_error(error))
    }
}

fn convert_palette_based(input: Input, _: Output, _: ColorType) -> Result<(), Error> {
    match png::Decoder::new(input).read_info() {
        Ok(mut reader) => {
            let mut buf  = vec![0; reader.output_buffer_size()];

            loop {
                match reader.next_frame(&mut buf) {
                    Ok(_) => {
                        println!("Frame: {}", buf.len());
                        for byte in &buf {
                            print!("[{}] ", byte);
                        }
                        println!("");
                    }
                    Err(err) => {
                        println!("Wuff: {}", err);
                        break;
                    }
                }
            }

            let info = reader.info();
            println!("picture: {} - {}, ({:?}, {:?}) ", info.width, info.height, info.color_type, info.bit_depth);
            if let Some(pal) = &info.palette {
                println!("Size of PAL: {}", pal.len());
            }



            Err(Error::from_str("Under construction"))
        },
        Err(error) => Err(Error::from_error(error))
    }
}

pub fn convert(args: Arguments, input: Input, output: Output) -> Result<(), Error> {
    match args.color_depth {
        ColorType::Full16 => convert_full16(input, output),
        _                 => convert_palette_based(input, output, args.color_depth),
    }
}